
//	Force Feedback Driver for Microsoft Xbox 360 Wireless Controller

#include	"effect.h"

//----------------------------------------------------------------------------------------------
//	CEffect
//----------------------------------------------------------------------------------------------
CEffect::CEffect()
{
	//	
	Type		= 0;
	//	DIEFFECT \
	ZeroMemory( &DiEffect, sizeof( DIEFFECT ) );
	//	DIENVELOPE \
	ZeroMemory( &DiEnvelope, sizeof( DIENVELOPE ) );
	//	DICONSTANTFORCE \
	ZeroMemory( &DiConstantForce, sizeof( DICONSTANTFORCE ) );
	//	DIPERIODIC \
	ZeroMemory( &DiPeriodic, sizeof( DIPERIODIC ) );
	//	DIRAMPFORCE \
	ZeroMemory( &DiRampforce, sizeof( DIRAMPFORCE ) );

	//	nh
	Handle		= NULL;

	//	
	Status		= 0;
	//	Đ
	PlayCount	= 0;
	//	Đ
	StartTime	= 0;
}

//----------------------------------------------------------------------------------------------
//	Calc
//----------------------------------------------------------------------------------------------
VOID CEffect::Calc(
	 LONG *	LeftLevel
	,LONG *	RightLevel )
{
	//	GtFNg̍ĐԁAJnAIAݎZo
	ULONG	Duration	= DiEffect.dwDuration / 1000;
	ULONG	BeginTime	= StartTime + ( DiEffect.dwStartDelay / 1000 );
	ULONG	EndTime		= 0xFFFFFFFF;
	if( PlayCount != -1 )
	{
		EndTime	= BeginTime + Duration * PlayCount;
	}
	ULONG	CurrentTime	= GetTickCount();

	//	GtFNg͍ĐH
	if( Status == DIEGES_PLAYING && BeginTime <= CurrentTime && CurrentTime <= EndTime )
	{
		//	Gx[vvZ
		LONG	NormalRate;
		LONG	AttackLevel;
		LONG	FadeLevel;
		CalcEnvelope(
			 Duration
			,( CurrentTime - BeginTime ) % Duration
			,&NormalRate
			,&AttackLevel
			,&FadeLevel );

		//	tH[XvZ
		LONG	NormalLevel;
		LONG	WorkLeftLevel;
		LONG	WorkRightLevel;
		CalcForce(
			 Duration
			,( CurrentTime - BeginTime ) % Duration
			,NormalRate
			,AttackLevel
			,FadeLevel
			,&NormalLevel );

		//	tH[X̐𒲐
		WorkLeftLevel	= (NormalLevel > 0) ? NormalLevel : -NormalLevel;
		WorkRightLevel	= (NormalLevel > 0) ? NormalLevel : -NormalLevel;

		//	tH[X̏𒲐
		WorkLeftLevel	= min( 65535, WorkLeftLevel * 65535 / 10000 );
		WorkRightLevel	= min( 65535, WorkRightLevel * 65535 / 10000 );

		//	tH[XZ
		*LeftLevel	= *LeftLevel + WorkLeftLevel;
		*RightLevel	= *RightLevel + WorkRightLevel;
	}
}

//----------------------------------------------------------------------------------------------
//	CalcEnvelope
//----------------------------------------------------------------------------------------------
VOID CEffect::CalcEnvelope(
	 ULONG	Duration
	,ULONG	CurrentPos
	,LONG *	NormalRate
	,LONG *	AttackLevel
	,LONG *	FadeLevel )
{
	//	Gx[vvZ
	if( ( DiEffect.dwFlags & DIEP_ENVELOPE ) && DiEffect.lpEnvelope != NULL )
	{
		//	A^bN̊Zo
		LONG	AttackRate	= 0;
		ULONG	AttackTime	= DiEnvelope.dwAttackTime / 1000;
		if( CurrentPos < AttackTime )
		{
			AttackRate	= ( AttackTime - CurrentPos ) * 100 / AttackTime;
		}
		//	tF[h̊Zo
		LONG	FadeRate	= 0;
		ULONG	FadeTime	= DiEnvelope.dwFadeTime / 1000;
		ULONG	FadePos		= Duration - FadeTime;
		if( FadePos < CurrentPos )
		{
			FadeRate	= ( CurrentPos - FadePos ) * 100 / FadeTime;
		}
		//	ZolԂ
		*NormalRate		= 100 - AttackRate - FadeRate;
		*AttackLevel	= DiEnvelope.dwAttackLevel * AttackRate;
		*FadeLevel		= DiEnvelope.dwFadeLevel * FadeRate;
	} else {
		*NormalRate		= 100;
		*AttackLevel	= 0;
		*FadeLevel		= 0;
	}
}

//----------------------------------------------------------------------------------------------
//	CalcForce
//----------------------------------------------------------------------------------------------
VOID CEffect::CalcForce(
	 ULONG		Duration
	,ULONG		CurrentPos
	,LONG		NormalRate
	,LONG		AttackLevel
	,LONG		FadeLevel
	,LONG *		NormalLevel )
{
	//	ϐ錾
	LONG	Magnitude	= 0;
	LONG	Period;
	LONG	R;
	LONG	Rate;

	//	GtFNg̎ނɂďU蕪
	switch( Type )
	{
		//	
		case SPRING:
		case DAMPER:
		case INERTIA:
		case FRICTION:
			;
			break;

		//	RX^g tH[X
		case CONSTANT_FORCE:
			//	Gx[vl}Oj`[h߂
			Magnitude	= DiConstantForce.lMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			break;

		//	JX^ tH[X
		case CUSTOM_FORCE:
			;
			break;

		//	IGtFNg
		case SQUARE:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( DiPeriodic.dwPeriod / 1000 ) );
			R		= ( CurrentPos % Period ) * 360 / Period;
			//	tF[YvZ
			R	= ( R + ( DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	`l}Oj`[h߂
			if( 180 <= R )
			{
				Magnitude	= Magnitude * -1;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + DiPeriodic.lOffset;
			break;

		case SINE:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( DiPeriodic.dwPeriod / 1000 ) );
			R		= ( CurrentPos % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	gl}Oj`[h߂
			Magnitude	= (LONG)( Magnitude * sin( R * 3.1415 / 180.0 ) );
			//	ItZbgvZ
			Magnitude	= Magnitude + DiPeriodic.lOffset;
			break;

		case TRIANGLE:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( DiPeriodic.dwPeriod / 1000 ) );
			R		= ( CurrentPos % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	Opgl}Oj`[h߂
			if( 0 <= R && R < 90 )
			{
				Magnitude	= -Magnitude * ( 90 - R ) / 90;
			}
			if( 90 <= R && R < 180 )
			{
				Magnitude	= Magnitude * ( R - 90 ) / 90;
			}
			if( 180 <= R && R < 270 )
			{
				Magnitude	= Magnitude * ( 90 - ( R - 180 ) ) / 90;
			}
			if( 270 <= R && R < 360 )
			{
				Magnitude	= -Magnitude * ( R - 270 ) / 90;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + DiPeriodic.lOffset;
			break;

		case SAWTOOTH_UP:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( DiPeriodic.dwPeriod / 1000 ) );
			R		= ( CurrentPos % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	Abvl}Oj`[h߂
			if( 0 <= R && R < 180 )
			{
				Magnitude	= -Magnitude * ( 180 - R ) / 180;
			}
			if( 180 <= R && R < 360 )
			{
				Magnitude	= Magnitude * ( R - 180 ) / 180;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + DiPeriodic.lOffset;
			break;

		case SAWTOOTH_DOWN:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( DiPeriodic.dwPeriod / 1000 ) );
			R		= ( CurrentPos % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	_El}Oj`[h߂
			if( 0 <= R && R < 180 )
			{
				Magnitude	= Magnitude * ( 180 - R ) / 180;
			}
			if( 180 <= R && R < 360 )
			{
				Magnitude	= -Magnitude * ( R - 180 ) / 180;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + DiPeriodic.lOffset;
			break;

		//	X΃tH[X
		case RAMP_FORCE:
			//	n_AI_̊Zo
			Rate		= ( Duration - CurrentPos ) * 100 / ( DiEffect.dwDuration / 1000 );
			//	Gx[vl}Oj`[h߂
			Magnitude	= ( DiRampforce.lStart * Rate
							+ DiRampforce.lEnd * ( 100 - Rate ) ) / 100;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			break;
	}

	//	QCltH[XԂ
	*NormalLevel	= Magnitude * (LONG)DiEffect.dwGain / 10000;
}
