
//	Force Feedback Driver for Microsoft Xbox 360 Wireless Controller

#include	"effect_driver.h"

//----------------------------------------------------------------------------------------------
//	CEffectDriver
//----------------------------------------------------------------------------------------------
CEffectDriver::CEffectDriver( VOID )
{
	//	QƃJE^
	ReferenceCount	= 1;

	//	foCX C^[tFCX
	Device			= NULL;

	//	[J Xbh
	WorkerThread	= NULL;

	//	GtFNgԍ
	EffectIndex		= 0;
	//	GtFNg
	EffectCount		= 0;
	//	GtFNg Xg
	EffectList		= NULL;

	//	~
	Stopped			= TRUE;
	//	ꎞ~
	Paused			= FALSE;
	//	ꎞ~
	PausedTime		= 0;

	//	QC
	Gain			= 10000;
	//	AN`G[^
	Actuator		= TRUE;

	//	IVOi
	Quit			= FALSE;
}

//----------------------------------------------------------------------------------------------
//	QueryInterface
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::QueryInterface(
	 REFIID		InterfaceID
	,PVOID *	Interface )
{
	//	ϐ錾
	HRESULT	Result	= S_OK;

	//	g̃C^[tFCXԂ
	if(	IsEqualIID( InterfaceID, IID_IUnknown )
	||	IsEqualIID( InterfaceID, IID_IDirectInputEffectDriver ) )
	{
		AddRef();
		*Interface	= this;
	} else {
		*Interface	= NULL;
		Result		= E_NOINTERFACE;
	}

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	AddRef
//----------------------------------------------------------------------------------------------
ULONG STDMETHODCALLTYPE CEffectDriver::AddRef( VOID )
{
	//	QƃJE^Z
	return( InterlockedIncrement( &ReferenceCount ) );
}

//----------------------------------------------------------------------------------------------
//	Release
//----------------------------------------------------------------------------------------------
ULONG STDMETHODCALLTYPE CEffectDriver::Release( VOID )
{
	//	QƃJE^Z
	if( InterlockedDecrement( &ReferenceCount ) == 0 )
	{
		//	hCȍԂIԂɂ
		Quit	= TRUE;

		//	[J[ XbhÎ҂
		for(;;)
		{
			DWORD	ExitCode;
			GetExitCodeThread( WorkerThread, &ExitCode );
			if( ExitCode != STILL_ACTIVE )
			{
				//	Xbh
				CloseHandle( WorkerThread );
				break;
			}
			//	1ms ҂
			Sleep( 1 );
		}

		//	foCX C^[tFCX
		if( Device != NULL )
		{
			CloseHandle( Device );
		}

		delete this;
	}

	return( ReferenceCount );
}

//----------------------------------------------------------------------------------------------
//	DeviceID
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::DeviceID(
	 DWORD	DiVersion
	,DWORD	ExternalID
	,DWORD	Begin
	,DWORD	InternalID
	,LPVOID	DiHIDFFInitInfo )
{
	//	foCX C^[tFCXJ
	Device	= CreateFile(
				 ((LPDIHIDFFINITINFO)DiHIDFFInitInfo)->pwszDeviceInterface
				,GENERIC_READ | GENERIC_WRITE
				,FILE_SHARE_READ | FILE_SHARE_WRITE
				,NULL
				,OPEN_EXISTING
				,FILE_ATTRIBUTE_NORMAL
				,NULL );
	if( Device == INVALID_HANDLE_VALUE )
	{
		return( E_FAIL );
	}

	//	[J[ XbhJn
	DWORD	ThreadID;
	WorkerThread	= CreateThread(
						 NULL
						,0
						,EffectProc
						,(LPVOID)this
						,0
						,&ThreadID );
	if( WorkerThread == NULL )
	{
		return( E_FAIL );
	}

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	GetVersions
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::GetVersions( LPDIDRIVERVERSIONS DriverVersions )
{
	//	hCo o[W̃TCY`FbN
	if( DriverVersions->dwSize != sizeof( DIDRIVERVERSIONS ) )
	{
		return( E_INVALIDARG );
	}

	//	hCo o[WԂ
	DriverVersions->dwFirmwareRevision	= 1;
	DriverVersions->dwHardwareRevision	= 1;
	DriverVersions->dwFFDriverVersion	= 1;

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	Escape
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::Escape(
	 DWORD			ExternalID
	,DWORD			Effect
	,LPDIEFFESCAPE	DIEffEscape )
{
	//	T|[g
	return( E_NOTIMPL );
}

//----------------------------------------------------------------------------------------------
//	SetGain
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::SetGain(
	 DWORD	ExternalID
	,DWORD	NewGain )
{
	//	ϐ錾
	HRESULT	Result	= S_OK;

	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	QCݒ肷
	if( 1 <= NewGain && NewGain <= 10000 )
	{
		Gain	= NewGain;
	} else {
		Gain	= max( 1, min( NewGain, 10000 ) );
		Result	= DI_TRUNCATED;
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	SendForceFeedbackCommand
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::SendForceFeedbackCommand(
	 DWORD	ExternalID
	,DWORD	Command )
{
	//	ϐ錾
	HRESULT	Result	= S_OK;

	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	R}hɂďU蕪
	switch( Command ) {

		case DISFFC_RESET:
			//	SẴGtFNg폜
			for( LONG Index = 0; Index < EffectCount; Index ++ )
			{
				delete EffectList[Index];
			}
			//	GtFNg 0 ɂ
			EffectCount	= 0;
			//	GtFNg Xg폜
			free( EffectList );
			EffectList	= NULL;
			//	Đ~
			Stopped	= TRUE;
			//	ꎞ~
			Paused	= FALSE;
			break;

		case DISFFC_STOPALL:
			//	SẴGtFNg̍Đ~
			for( LONG Index = 0; Index < EffectCount; Index ++ )
			{
				EffectList[Index]->Status	= NULL;
			}
			//	Đ~
			Stopped	= TRUE;
			//	ꎞ~
			Paused	= FALSE;
			break;

		case DISFFC_PAUSE:
			//	Đꎞ~
			Paused		= TRUE;
			PausedTime	= GetTickCount();
			break;

		case DISFFC_CONTINUE:
			//	ꎞ~
			for( LONG Index = 0; Index < EffectCount; Index ++ )
			{
				//	ꎞ~ԂGtFNg̊JnԂx点
				EffectList[Index]->StartTime	+= ( GetTickCount() - PausedTime );
			}
			Paused	= FALSE;
			break;

		case DISFFC_SETACTUATORSON:
			//	AN`G[^Lɂ
			Actuator	= TRUE;
			break;

		case DISFFC_SETACTUATORSOFF:
			//	AN`G[^𖳌ɂ
			Actuator	= FALSE;
			break;

		default:
			Result	= E_NOTIMPL;
			break;
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	GetForceFeedbackState
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::GetForceFeedbackState(
	 DWORD				ExternalID
	,LPDIDEVICESTATE	DeviceState )
{
	//	foCXԂ̃TCY`FbN
	if( DeviceState->dwSize != sizeof( DIDEVICESTATE ) )
	{
		return( E_INVALIDARG );
	}

	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	GtFNg hCȍԂԂ
	DeviceState->dwState	= NULL;
	//	GtFNg Xg󂩂ǂ
	if( EffectCount == 0 )
	{
		DeviceState->dwState	|= DIGFFS_EMPTY;
	}
	//	GtFNg~ǂ
	if( Stopped == TRUE )
	{
		DeviceState->dwState	|= DIGFFS_STOPPED;
	}
	//	GtFNgꎞ~ǂ
	if( Paused == TRUE )
	{
		DeviceState->dwState	|= DIGFFS_PAUSED;
	}
	//	AN`G[^쓮Ă邩ǂ
	if( Actuator == TRUE )
	{
		DeviceState->dwState	|= DIGFFS_ACTUATORSON;
	} else {
		DeviceState->dwState	|= DIGFFS_ACTUATORSOFF;
	}
	//	d̓IŒ
	DeviceState->dwState	|= DIGFFS_POWERON;
	//	Z[teB XCb`̓ItŒ
	DeviceState->dwState	|= DIGFFS_SAFETYSWITCHOFF;
	//	[U[ XCb`̓IŒ
	DeviceState->dwState	|= DIGFFS_USERFFSWITCHON;

	//	[h\͖Œ
	DeviceState->dwLoad		= 0;

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	DownloadEffect
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::DownloadEffect(
	 DWORD			ExternalID
	,DWORD			EffectType
	,LPDWORD		EffectHandle
	,LPCDIEFFECT	DiEffect
	,DWORD			Flags )
{
	//	ϐ錾
	CEffect *	Effect;

	//	DIEP_NODOWNLOAD ̏ꍇ͖
	if( Flags & DIEP_NODOWNLOAD )
	{
		return( S_OK );
	}

	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	if( *EffectHandle == NULL )
	{
		//	VKGtFNg쐬
		Effect	= new CEffect();
		Effect->Handle	= ( EffectIndex ++ );
		//	GtFNg𑝂₷
		EffectCount	++;
		//	GtFNg Xgg
		EffectList	= (CEffect * *)realloc( EffectList, sizeof( CEffect * ) * EffectCount );
		EffectList[EffectCount - 1]	= Effect;
		//	GtFNg̃nhԂ
		*EffectHandle	= Effect->Handle;
	} else {
		//	nhGtFNg
		for( LONG Index = 0; Index < EffectCount; Index ++ )
		{
			if( EffectList[Index]->Handle == *EffectHandle )
			{
				Effect	= EffectList[Index];
				break;
			}
		}
	}

	//	ނݒ肷
	Effect->Type	= EffectType;

	//	Flags ݒ肷
	Effect->DiEffect.dwFlags	= DiEffect->dwFlags;

	//	DIEP_DURATION ݒ肷
	if( Flags & DIEP_DURATION )
	{
		Effect->DiEffect.dwDuration	= DiEffect->dwDuration;
	}

	//	DIEP_SAMPLEPERIOD ݒ肷
	if( Flags & DIEP_SAMPLEPERIOD )
	{
		Effect->DiEffect.dwSamplePeriod	= DiEffect->dwSamplePeriod;
	}

	//	DIEP_GAIN ݒ肷
	if( Flags & DIEP_GAIN )
	{
		Effect->DiEffect.dwGain	= DiEffect->dwGain;
	}

	//	DIEP_TRIGGERBUTTON ݒ肷
	if( Flags & DIEP_TRIGGERBUTTON )
	{
		Effect->DiEffect.dwTriggerButton	= DiEffect->dwTriggerButton;
	}

	//	DIEP_TRIGGERREPEATINTERVAL ݒ肷
	if( Flags & DIEP_TRIGGERREPEATINTERVAL )
	{
		Effect->DiEffect.dwTriggerRepeatInterval	= DiEffect->dwTriggerRepeatInterval;
	}

	//	DIEP_AXES ݒ肷
	if( Flags & DIEP_AXES )
	{
		Effect->DiEffect.cAxes		= DiEffect->cAxes;
		Effect->DiEffect.rgdwAxes	= NULL;
	}

	//	DIEP_DIRECTION ݒ肷
	if( Flags & DIEP_DIRECTION )
	{
		Effect->DiEffect.cAxes			= DiEffect->cAxes;
		Effect->DiEffect.rglDirection	= NULL;
	}

	//	DIEP_ENVELOPE ݒ肷
	if( ( Flags & DIEP_ENVELOPE ) && DiEffect->lpEnvelope != NULL )
	{
		CopyMemory( &Effect->DiEnvelope, DiEffect->lpEnvelope, sizeof( DIENVELOPE ) );
		//	A^bNԂƃtF[hԂԂꍇ͒l␳
		if( Effect->DiEffect.dwDuration - Effect->DiEnvelope.dwFadeTime
				< Effect->DiEnvelope.dwAttackTime )
		{
			Effect->DiEnvelope.dwFadeTime	= Effect->DiEnvelope.dwAttackTime;
		}
		Effect->DiEffect.lpEnvelope	= &Effect->DiEnvelope;
	}

	//	TypeSpecificParams ݒ肷
	Effect->DiEffect.cbTypeSpecificParams	= DiEffect->cbTypeSpecificParams;

	//	DIEP_TYPESPECIFICPARAMS ݒ肷
	if( Flags & DIEP_TYPESPECIFICPARAMS )
	{
		switch( EffectType )
		{
			case SPRING:
			case DAMPER:
			case INERTIA:
			case FRICTION:
				Effect->DiEffect.lpvTypeSpecificParams	= NULL;
				break;

			case CONSTANT_FORCE:
				CopyMemory(
					 &Effect->DiConstantForce
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= &Effect->DiConstantForce;
				break;

			case CUSTOM_FORCE:
				Effect->DiEffect.lpvTypeSpecificParams	= NULL;
				break;

			case SQUARE:
			case SINE:
			case TRIANGLE:
			case SAWTOOTH_UP:
			case SAWTOOTH_DOWN:
				CopyMemory(
					 &Effect->DiPeriodic
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= &Effect->DiPeriodic;
				break;

			case RAMP_FORCE:
				CopyMemory(
					 &Effect->DiRampforce
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= &Effect->DiRampforce;
				break;
		}
	}

	//	DIEP_STARTDELAY ݒ肷
	if( Flags & DIEP_STARTDELAY )
	{
		Effect->DiEffect.dwStartDelay	= DiEffect->dwStartDelay;
	}

	//	DIEP_START w肳Ă΃GtFNgĐ
	if( Flags & DIEP_START )
	{
		Effect->Status		= DIEGES_PLAYING;
		Effect->PlayCount	= 1;
		Effect->StartTime	= GetTickCount();
	}

	//	DIEP_NORESTART ͖
	if( Flags & DIEP_NORESTART )
	{
		;
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	DestroyEffect
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::DestroyEffect(
	 DWORD	ExternalID
	,DWORD	EffectHandle )
{
	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	nhGtFNg
	for( LONG Index = 0; Index < EffectCount; Index ++ )
	{
		if( EffectList[Index]->Handle == EffectHandle )
		{
			//	GtFNg폜
			delete EffectList[Index];
			//	GtFNg炷
			EffectCount	--;
			//	GtFNg Xgk
			CopyMemory(
				 &EffectList[Index]
				,&EffectList[Index + 1]
				,sizeof( CEffect * ) * ( EffectCount - Index ) );
			EffectList	= (CEffect * *)realloc( EffectList, sizeof( CEffect * ) * EffectCount );
			break;
		}
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	StartEffect
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::StartEffect(
	 DWORD	ExternalID
	,DWORD	EffectHandle
	,DWORD	Mode
	,DWORD	Count )
{
	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	nhGtFNg
	for( LONG Index = 0; Index < EffectCount; Index ++ )
	{
		if( EffectList[Index]->Handle == EffectHandle )
		{
			//	GtFNg̍ĐJn
			EffectList[Index]->Status		= DIEGES_PLAYING;
			EffectList[Index]->PlayCount	= Count;
			EffectList[Index]->StartTime	= GetTickCount();
			//	~
			Stopped	= FALSE;
		} else {
			//	DIES_SOLO w肳Ăꍇ͑̃GtFNg~
			if( Mode & DIES_SOLO )
			{
				EffectList[Index]->Status	= NULL;
			}
		}
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	StopEffect
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::StopEffect(
	 DWORD	ExternalID
	,DWORD	EffectHandle )
{
	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	nhGtFNg
	for( LONG Index = 0; Index < EffectCount; Index ++ )
	{
		if( EffectList[Index]->Handle == EffectHandle )
		{
			//	GtFNg~
			EffectList[Index]->Status	= NULL;
			break;
		}
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	GetEffectStatus
//----------------------------------------------------------------------------------------------
HRESULT STDMETHODCALLTYPE CEffectDriver::GetEffectStatus(
	 DWORD		ExternalID
	,DWORD		EffectHandle
	,LPDWORD	Status )
{
	//	NeBJ ZNVɓ
	EnterCriticalSection( &CriticalSection );

	//	nhGtFNg
	for( LONG Index = 0; Index < EffectCount; Index ++ )
	{
		if( EffectList[Index]->Handle == EffectHandle )
		{
			//	GtFNg̏ԂԂ
			*Status	= EffectList[Index]->Status;
			break;
		}
	}

	//	NeBJ ZNVo
	LeaveCriticalSection( &CriticalSection );

	return( S_OK );
}

//----------------------------------------------------------------------------------------------
//	EffectProc
//----------------------------------------------------------------------------------------------
STDAPI_(DWORD) WINAPI EffectProc( LPVOID EffectDriverInterface )
{
	//	ϐ錾
	HID_WRITE_REPORT_ID	Report;
	DWORD				ReqLen;

	//	GtFNg hCo擾
	CEffectDriver *	EffectDriver	= (CEffectDriver *)EffectDriverInterface;

	//	ÕGtFNg̋
	LONG	PrvLeftLevel	= -1;
	LONG	PrvRightLevel	= -1;

	while( EffectDriver->Quit == FALSE )
	{
		//	NeBJ ZNVɓ
		EnterCriticalSection( &CriticalSection );

		//	GtFNg̋
		LONG	LeftLevel	= 0;
		LONG	RightLevel	= 0;
		LONG	Gain		= EffectDriver->Gain;

		if( EffectDriver->Actuator == TRUE )
		{
			//	GtFNg̋vZ
			for( LONG Index = 0; Index < EffectDriver->EffectCount; Index ++ )
			{
				EffectDriver->EffectList[Index]->Calc( &LeftLevel, &RightLevel );
			}
		}

		//	NeBJ ZNVo
		LeaveCriticalSection( &CriticalSection );

		//	Rg[[̐Uݒ肷
		if( PrvLeftLevel != LeftLevel || PrvRightLevel != RightLevel )
		{
			//	foCXփtH[Xo
			Report.Id		= 0;
			Report.Type		= REPORT_TYPE_NORMAL;
			Report.Left		= (UCHAR)min( 255, ( LeftLevel * Gain / 10000 ) >> 8 );
			Report.Right	= (UCHAR)min( 255, ( LeftLevel * Gain / 10000 ) >> 8 );
			WriteFile(
				 EffectDriver->Device
				,&Report
				,sizeof( HID_WRITE_REPORT_ID )
				,&ReqLen
				,NULL );
		}

		//	GtFNg̋ޔ
		PrvLeftLevel	= LeftLevel;
		PrvRightLevel	= RightLevel;

		//	2ms ҋ@
		Sleep( 2 );
	}

	//	foCXփtH[Xo
	Report.Id		= 0;
	Report.Type		= REPORT_TYPE_NORMAL;
	Report.Left		= 0;
	Report.Right	= 0;
	WriteFile( EffectDriver->Device, &Report, sizeof( HID_WRITE_REPORT_ID ), &ReqLen, NULL );

	return( 0 );
}
