
//	HID Minidriver for Microsoft Xbox 360 Wireless Controller

#include	"driver.h"

//----------------------------------------------------------------------------------------------
//	֐
//----------------------------------------------------------------------------------------------

NTSTATUS IoCompletionRoutine( IN PDEVICE_OBJECT, IN PIRP, IN PKEVENT );

//----------------------------------------------------------------------------------------------
//	CompleteRequest
//----------------------------------------------------------------------------------------------
VOID CompleteRequest(
	 IN PIRP		Irp
	,IN NTSTATUS	Status
	,IN ULONG		Information )
{
	//	I/O v
	Irp->IoStatus.Status		= Status;
	Irp->IoStatus.Information	= Information;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
}

//----------------------------------------------------------------------------------------------
//	SendRequest
//----------------------------------------------------------------------------------------------
NTSTATUS SendRequest(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ʃhCo I/O v𔭍s
	IoSkipCurrentIrpStackLocation( Irp );
	return( IoCallDriver( GET_NEXT_DEVICE_OBJECT( DeviceObject ), Irp ) );
}

//----------------------------------------------------------------------------------------------
//	SendAndWaitRequest
//----------------------------------------------------------------------------------------------
NTSTATUS SendAndWaitRequest(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ϐ錾
	NTSTATUS	Status;
	KEVENT		WaitEvent;

	//	ʃhCo I/O v𔭍s
	KeInitializeEvent( &WaitEvent, NotificationEvent, FALSE );
	IoCopyCurrentIrpStackLocationToNext( Irp );
	IoSetCompletionRoutine(
		 Irp
		,(PIO_COMPLETION_ROUTINE)IoCompletionRoutine
		,(PVOID)&WaitEvent
		,TRUE
		,TRUE
		,TRUE );
	Status	= IoCallDriver( GET_NEXT_DEVICE_OBJECT( DeviceObject ), Irp );

	//	ʃhCȍ܂őҋ@
	if( Status == STATUS_PENDING )
	{
		KeWaitForSingleObject( &WaitEvent, Executive, KernelMode, FALSE, NULL );
		Status	= Irp->IoStatus.Status;
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	SendPowerRequest
//----------------------------------------------------------------------------------------------
NTSTATUS SendPowerRequest(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ʃhCo I/O v𔭍s
	PoStartNextPowerIrp( Irp );
	IoSkipCurrentIrpStackLocation( Irp );
	return( PoCallDriver( GET_NEXT_DEVICE_OBJECT( DeviceObject ), Irp ) );
}

//----------------------------------------------------------------------------------------------
//	SendAndWaitPowerRequest
//----------------------------------------------------------------------------------------------
NTSTATUS SendAndWaitPowerRequest(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ϐ錾
	NTSTATUS	Status;
	KEVENT		WaitEvent;

	//	ʃhCo I/O v𔭍s
	KeInitializeEvent( &WaitEvent, NotificationEvent, FALSE );
	PoStartNextPowerIrp( Irp );
	IoCopyCurrentIrpStackLocationToNext( Irp );
	IoSetCompletionRoutine(
		 Irp
		,(PIO_COMPLETION_ROUTINE)IoCompletionRoutine
		,(PVOID)&WaitEvent
		,TRUE
		,TRUE
		,TRUE );
	Status	= PoCallDriver( GET_NEXT_DEVICE_OBJECT( DeviceObject ), Irp );

	//	ʃhCȍ܂őҋ@
	if( Status == STATUS_PENDING )
	{
		KeWaitForSingleObject( &WaitEvent, Executive, KernelMode, FALSE, NULL );
		Status	= Irp->IoStatus.Status;
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IoCompletionRoutine
//----------------------------------------------------------------------------------------------
NTSTATUS IoCompletionRoutine(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp
	,IN PKEVENT			Event )
{
	//	I/O v̊ʒm
	KeSetEvent( Event, 0, FALSE );

	return( STATUS_MORE_PROCESSING_REQUIRED );
}

//----------------------------------------------------------------------------------------------
//	IncrementReference
//----------------------------------------------------------------------------------------------
NTSTATUS IncrementReference( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	QƃJE^Z
	InterlockedIncrement( &DeviceExtension->ReferenceCount );

	//	hCȍԂ͗LEH
	if( DeviceExtension->DriverState == Disable )
	{
		if( InterlockedDecrement( &DeviceExtension->ReferenceCount ) == 0 )
		{
			//	QƏICxg𑗂
			KeSetEvent( &DeviceExtension->ReferenceEnd, 0, FALSE );
			Status	= STATUS_DELETE_PENDING;
		}
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	DecrementReference
//----------------------------------------------------------------------------------------------
VOID DecrementReference( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	QƃJE^Z
	if( InterlockedDecrement( &DeviceExtension->ReferenceCount ) == 0 )
	{
		//	QƏICxg𑗂
		KeSetEvent( &DeviceExtension->ReferenceEnd, 0, FALSE );
	}
}

//----------------------------------------------------------------------------------------------
//	GetRegistryBinary
//----------------------------------------------------------------------------------------------
VOID GetRegistryBinary(
	 IN PWSTR	Path
	,IN PWSTR	Value
	,IN CHAR *	Buf
	,IN LONG	BufSize )
{
	//	ϐ錾
	NTSTATUS						Status;
	UNICODE_STRING					RegistryPath;
	UNICODE_STRING					RegistryValue;
	OBJECT_ATTRIBUTES				ObjAttribute;
	HANDLE							Key;
	ULONG							ValueSize	= 0;
	PKEY_VALUE_PARTIAL_INFORMATION	ValueInfo;

	//	WXg pX쐬
	RtlInitUnicodeString( &RegistryPath, Path );
	//	IuWFNg쐬
	InitializeObjectAttributes(
		 &ObjAttribute
		,&RegistryPath
		,OBJ_CASE_INSENSITIVE
		,NULL
		,NULL );

	//	WXgJ
	Status	= ZwOpenKey( &Key, KEY_QUERY_VALUE, &ObjAttribute );
	if( NT_SUCCESS( Status ) )
	{
		//	WXgl쐬
		RtlInitUnicodeString( &RegistryValue, Value );

		//	WXgl̃TCY擾
		Status	= ZwQueryValueKey(
					 Key
					,&RegistryValue
					,KeyValuePartialInformation
					,NULL
					,0
					,&ValueSize );
		if( ( Status == STATUS_OBJECT_NAME_NOT_FOUND ) || ( ValueSize == 0 ) )
		{
			ZwClose( Key );
			return;
		}

		//	WXgl擾obt@쐬
		ValueInfo	= (PKEY_VALUE_PARTIAL_INFORMATION)ExAllocatePool( NonPagedPool, ValueSize );
		if( ValueInfo == NULL )
		{
			ZwClose( Key );
			return;
		}

		//	WXgl擾
		Status	= ZwQueryValueKey(
					 Key
					,&RegistryValue
					,KeyValuePartialInformation
					,ValueInfo
					,ValueSize
					,&ValueSize );
		if( !NT_SUCCESS( Status ) )
		{
			ExFreePool( ValueInfo );
			ZwClose( Key );
			return;
		}

		//	擾eޔ
		RtlCopyMemory( Buf, ValueInfo->Data, BufSize );

		//	obt@J
		ExFreePool( ValueInfo );

		//	WXg
		ZwClose( Key );
	}
}
