
//	HID Minidriver for Microsoft Xbox 360 Wireless Controller

#include	"driver.h"

//----------------------------------------------------------------------------------------------
//	萔
//----------------------------------------------------------------------------------------------

#define	INTERVAL					-1
#define	INTERVAL_STALL				-16 * 1000 * 10

//----------------------------------------------------------------------------------------------
//	֐
//----------------------------------------------------------------------------------------------

NTSTATUS StartDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS QueryCapabilities( IN PDEVICE_OBJECT, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS SurpriseRemoval( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS RemoveDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
VOID WorkerThreadIn( IN PVOID );
VOID WorkerThreadOut( IN PVOID );

//----------------------------------------------------------------------------------------------
//	PnP
//----------------------------------------------------------------------------------------------
NTSTATUS PnP(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ϐ錾
	NTSTATUS			Status;
	PDEVICE_EXTENSION	DeviceExtension	= GET_MINIDRIVER_DEVICE_EXTENSION( DeviceObject );
	PIO_STACK_LOCATION	CurrentIrpStack;

	//	QƃJE^Z
	Status	= IncrementReference( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		CompleteRequest( Irp, Status, 0 );
		return( Status );
	}

	//	}Ci[ t@NVɂďU蕪
	CurrentIrpStack	= IoGetCurrentIrpStackLocation( Irp );
	switch( CurrentIrpStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE:
			Status	= StartDevice( DeviceObject, DeviceExtension, Irp );
			break;

		case IRP_MN_QUERY_CAPABILITIES:
			Status	= QueryCapabilities( DeviceObject, Irp, CurrentIrpStack );
			break;

		case IRP_MN_SURPRISE_REMOVAL:
			Status	= SurpriseRemoval( DeviceObject, DeviceExtension, Irp );
			break;

		case IRP_MN_REMOVE_DEVICE:
			Status	= RemoveDevice( DeviceObject, DeviceExtension, Irp );
			break;

		default:
			//	IRP_MJ_PNP vʃhCoɑ
			Status	= SendRequest( DeviceObject, Irp );
			break;
	}

	//	QƃJE^Z
	DecrementReference( DeviceExtension );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	StartDevice
//----------------------------------------------------------------------------------------------
NTSTATUS StartDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	ϐ錾
	NTSTATUS	Status;

	//	 IRP_MN_START_DEVICE vʃhCoɑ
	Status	= SendAndWaitRequest( DeviceObject, Irp );
	if( !NT_SUCCESS( Status ) )
	{
		CompleteRequest( Irp, Status, 0 );
		return( Status );
	}

	//	foCXLɂ
	Status	= EnableDevice( DeviceExtension );

	//	IRP_MN_START_DEVICE v
	CompleteRequest( Irp, Status, 0 );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	QueryCapabilities
//----------------------------------------------------------------------------------------------
NTSTATUS QueryCapabilities(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PIRP				Irp
	,IN PIO_STACK_LOCATION	CurrentIrpStack )
{
	//	ϐ錾
	NTSTATUS				Status;
	PDEVICE_CAPABILITIES	DeviceCapabilities;

	//	 IRP_MN_QUERY_CAPABILITIES vʃhCoɑ
	Status	= SendAndWaitRequest( DeviceObject, Irp );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	ʃhCoŐݒ肳ꂽfoCX̋@\ύX
	DeviceCapabilities	= CurrentIrpStack->Parameters.DeviceCapabilities.Capabilities;
	DeviceCapabilities->SurpriseRemovalOK	= TRUE;

	//	IRP_MN_QUERY_CAPABILITIES v
	CompleteRequest( Irp, Status, 0 );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	SurpriseRemoval
//----------------------------------------------------------------------------------------------
NTSTATUS SurpriseRemoval(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	foCX𖳌ɂ
	DisableDevice( DeviceExtension );

	//	IRP_MN_SURPRISE_REMOVAL vʃhCoɑ
	return( SendRequest( DeviceObject, Irp ) );
}

//----------------------------------------------------------------------------------------------
//	RemoveDevice
//----------------------------------------------------------------------------------------------
NTSTATUS RemoveDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	foCX𖳌ɂ
	DisableDevice( DeviceExtension );

	//	IRP_MN_REMOVE_DEVICE vʃhCoɑ
	return( SendRequest( DeviceObject, Irp ) );
}

//----------------------------------------------------------------------------------------------
//	EnableDevice
//----------------------------------------------------------------------------------------------
NTSTATUS EnableDevice( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status;
	HANDLE		ThreadHandle;

	//	USB |[g
	Status	= SetupPort( DeviceExtension->NextDeviceObject );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	foCXLqq擾
	Status	= GetDeviceDescriptor(
				 DeviceExtension->NextDeviceObject
				,&DeviceExtension->DeviceDescriptor );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	RtBM[VLqq擾
	Status	= GetConfigurationDescriptor(
				 DeviceExtension->NextDeviceObject
				,&DeviceExtension->ConfigurationDescriptor );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	C^[tFCXLqq擾
	Status	= GetInterfaceDescriptor(
				 DeviceExtension->ConfigurationDescriptor
				,&DeviceExtension->InterfaceDescriptor );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	RtBM[VI
	Status	= SelectConfiguration(
				 DeviceExtension->NextDeviceObject
				,DeviceExtension->ConfigurationDescriptor
				,DeviceExtension->InterfaceDescriptor
				,&DeviceExtension->InputPipeHandle
				,&DeviceExtension->OutputPipeHandle );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	ݒ
	DeviceExtension->ResetDevice	= TRUE;

	//	foCX̏ԂLɂ
	DeviceExtension->DeviceState	= Enable;

	//	͗p[J XbhJn
	DeviceExtension->ThreadState	= Enable;
	PsCreateSystemThread(
		 &ThreadHandle
		,THREAD_ALL_ACCESS
		,NULL
		,NULL
		,NULL
		,WorkerThreadIn
		,DeviceExtension );
	ObReferenceObjectByHandle(
		 ThreadHandle
		,THREAD_ALL_ACCESS
		,NULL
		,KernelMode
		,&DeviceExtension->ThreadObjectIn
		,NULL );
	ZwClose( ThreadHandle );

	//	o͗p[J XbhJn
	if( DeviceExtension->OutputPipeHandle != NULL )
	{
		PsCreateSystemThread(
			 &ThreadHandle
			,THREAD_ALL_ACCESS
			,NULL
			,NULL
			,NULL
			,WorkerThreadOut
			,DeviceExtension );
		ObReferenceObjectByHandle(
			 ThreadHandle
			,THREAD_ALL_ACCESS
			,NULL
			,KernelMode
			,&DeviceExtension->ThreadObjectOut
			,NULL );
		ZwClose( ThreadHandle );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	WorkerThreadIn
//----------------------------------------------------------------------------------------------
VOID WorkerThreadIn( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status;
	PDEVICE_EXTENSION	DeviceExtension	= Context;
	LARGE_INTEGER		TimeOut;
	INPUT_REPORT		ControllerReport;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	for(;;)
	{
		//	Xbh̏I`FbN
		if( DeviceExtension->ThreadState == Disable )
		{
			break;
		}

		//	Rg[[̐ݒǂݍ
		if(	DeviceExtension->ResetDevice		== TRUE
		&&	DeviceExtension->OutputPipeHandle	== NULL )
		{
			LoadSetting( &DeviceExtension->Setting );
			DeviceExtension->ResetDevice	= FALSE;
		}

		//	Rg[[ |[g̓͂s
		Status	= InputReport(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->InputPipeHandle
					,&ControllerReport
					,sizeof( INPUT_REPORT ) );
		if( NT_SUCCESS( Status ) )
		{
			//	Rg[[ڑꂽꍇ
			if(	ControllerReport.Type		== 0x08
			&&	ControllerReport.ParamType	== 0x80 )
			{
				DeviceExtension->Connected			= TRUE;
				DeviceExtension->ConnectionReported	= TRUE;
				DeviceExtension->ResetDevice		= TRUE;
			}
			//	Rg[[ؒfꂽꍇ
			if(	ControllerReport.Type		== 0x08
			&&	ControllerReport.ParamType	== 0x00 )
			{
				DeviceExtension->Connected			= FALSE;
				DeviceExtension->ConnectionReported	= TRUE;
			}
			//	͗p|[g擾ꍇ
			if(	ControllerReport.ParamType	== 0x01
			&&	ControllerReport.Param1		== 0xF0
			&&	DeviceExtension->Connected	== TRUE )
			{
				RtlCopyMemory(
					 &DeviceExtension->ControllerReport
					,&ControllerReport
					,sizeof( INPUT_REPORT ) );
			}
			TimeOut.QuadPart	= INTERVAL;
		} else {
			//	G[pCvZbg
			if( Status != STATUS_CANCELLED && Status != STATUS_DEVICE_NOT_CONNECTED )
			{
				ResetPipe(
					 DeviceExtension->NextDeviceObject
					,DeviceExtension->InputPipeHandle );
			}
			TimeOut.QuadPart	= INTERVAL_STALL;
		}
		//	ҋ@
		KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectIn );
	DeviceExtension->ThreadObjectIn	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingIn, 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS );
}

//----------------------------------------------------------------------------------------------
//	WorkerThreadOut
//----------------------------------------------------------------------------------------------
VOID WorkerThreadOut( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status;
	PDEVICE_EXTENSION	DeviceExtension	= Context;
	LARGE_INTEGER		TimeOut;
	POUTPUT_REPORT		ConnectionReport	= &DeviceExtension->ConnectionReport;
	POUTPUT_REPORT		RingLightReport		= &DeviceExtension->RingLightReport;
	POUTPUT_REPORT		ActuatorsReport		= &DeviceExtension->ActuatorsReport;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	for(;;)
	{
		//	Xbh̏I`FbN
		if( DeviceExtension->ThreadState == Disable )
		{
			break;
		}

		//	Rg[[̐ݒǂݍ
		if( DeviceExtension->ResetDevice == TRUE )
		{
			LoadSetting( &DeviceExtension->Setting );
			DeviceExtension->ActuatorsReported	= FALSE;
			DeviceExtension->RingLightReported	= FALSE;
			DeviceExtension->ResetDevice		= FALSE;
		}

		//	Rg[[̐ڑmF|[g̏o͂s
		if( DeviceExtension->ConnectionReported == FALSE )
		{
			//	Rg[[̐ڑmF|[g쐬
			ConnectionReport->Type		= 0x08;
			ConnectionReport->ParamType	= 0x00;
			ConnectionReport->Param1	= 0x0F;
			ConnectionReport->Param2	= 0xC0;
			//	Rg[[̐ڑmF|[go͂
			Status	= OutputReport(
						 DeviceExtension->NextDeviceObject
						,DeviceExtension->OutputPipeHandle
						,&DeviceExtension->ConnectionReport
						,sizeof( OUTPUT_REPORT ) );
			if( NT_SUCCESS( Status ) )
			{
				;
			}
		}

		//	O Cg |[g̏o͂s
		if(	DeviceExtension->RingLightReported	== FALSE
		&&	DeviceExtension->Connected			== TRUE )
		{
			//	O Cg |[g쐬
			RingLightReport->Type		= 0x00;
			RingLightReport->ParamType	= 0x00;
			RingLightReport->Param1		= 0x08;
			RingLightReport->Param2		= 0x40 + (UCHAR)DeviceExtension->Setting.RingLightPattern;
			//	O Cg |[go͂
			Status	= OutputReport(
						 DeviceExtension->NextDeviceObject
						,DeviceExtension->OutputPipeHandle
						,&DeviceExtension->RingLightReport
						,sizeof( OUTPUT_REPORT ) );
			if( NT_SUCCESS( Status ) )
			{
				DeviceExtension->RingLightReported	= TRUE;
			}
		}

		//	AN`G[^ |[g쐬
		if( DeviceExtension->Connected == TRUE )
		{
			ActuatorsReport->Type		= 0x00;
			ActuatorsReport->ParamType	= 0x01;
			ActuatorsReport->Param1		= 0x0F;
			ActuatorsReport->Param2		= 0xC0;
			if( DeviceExtension->ActuatorsReported == FALSE )
			{
				ActuatorsReport->LeftActuator	= 0;
				ActuatorsReport->RightActuator	= 0;
			}
			//	AN`G[^ |[go͂
			Status	= OutputReport(
						 DeviceExtension->NextDeviceObject
						,DeviceExtension->OutputPipeHandle
						,&DeviceExtension->ActuatorsReport
						,sizeof( OUTPUT_REPORT ) );
			if( NT_SUCCESS( Status ) )
			{
				DeviceExtension->ActuatorsReported	= TRUE;
				TimeOut.QuadPart					= INTERVAL;
			} else {
				//	G[pCvZbg
				if( Status != STATUS_CANCELLED && Status != STATUS_DEVICE_NOT_CONNECTED )
				{
					ResetPipe(
						 DeviceExtension->NextDeviceObject
						,DeviceExtension->OutputPipeHandle );
				}
				TimeOut.QuadPart	= INTERVAL_STALL;
			}
		} else {
			TimeOut.QuadPart	= INTERVAL_STALL;
		}
		//	ҋ@
		KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectOut );
	DeviceExtension->ThreadObjectOut	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingOut, 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS );
}

//----------------------------------------------------------------------------------------------
//	DisableDevice
//----------------------------------------------------------------------------------------------
NTSTATUS DisableDevice( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	[J XbhÎ҂
	if( DeviceExtension->ThreadState == Enable )
	{
		//	[J XbhI
		DeviceExtension->ThreadState	= Disable;

		//	͗ppCvA{[g
		AbortPipe( DeviceExtension->NextDeviceObject, DeviceExtension->InputPipeHandle );
		//	͗p[J Xbĥ҂
		KeWaitForSingleObject(
			 &DeviceExtension->ThreadExitingIn
			,Executive
			,KernelMode
			,FALSE
			,NULL );

		if( DeviceExtension->OutputPipeHandle != NULL )
		{
			//	o͗ppCvA{[g
			AbortPipe( DeviceExtension->NextDeviceObject, DeviceExtension->OutputPipeHandle );
			//	o͗p[J Xbĥ҂
			KeWaitForSingleObject(
				 &DeviceExtension->ThreadExitingOut
				,Executive
				,KernelMode
				,FALSE
				,NULL );
		}
	}

	//	hCȍԂ𖳌ɂ
	DeviceExtension->DriverState	= Disable;

	//	QƏICxg
	KeResetEvent( &DeviceExtension->ReferenceEnd );

	//	QƃJE^Z
	DecrementReference( DeviceExtension );
	DecrementReference( DeviceExtension );

	//	SĂ I/O vIAQƏIʒm܂őҋ@
	KeWaitForSingleObject( &DeviceExtension->ReferenceEnd, Executive, KernelMode, FALSE, NULL);

	//	foCX𖳌ɂ
	if( DeviceExtension->DeviceState == Enable )
	{
		//	RtBM[V̑I
		DeselectConfiguration( DeviceExtension->NextDeviceObject );

		//	RtBM[VLqqpobt@mۂĂΊJ
		if( DeviceExtension->ConfigurationDescriptor != NULL )
		{
			ExFreePool( DeviceExtension->ConfigurationDescriptor );
			DeviceExtension->ConfigurationDescriptor	= NULL;
		}

		//	C^[tFCXLqqpobt@mۂĂΊJ
		if( DeviceExtension->InterfaceDescriptor != NULL )
		{
			ExFreePool( DeviceExtension->InterfaceDescriptor );
			DeviceExtension->InterfaceDescriptor	= NULL;
		}

		//	foCX̏Ԃ𖳌ɂ
		DeviceExtension->DeviceState	= Disable;
	}

	//	hCȍԂLɂ
	DeviceExtension->DriverState	= Enable;

	//	QƃJE^̐킹
	IncrementReference( DeviceExtension );
	IncrementReference( DeviceExtension );

	return( STATUS_SUCCESS );
}
